;       Last change    August 1993
(in-package epilib)

;;;;   SORTS

(defvar *sorts* nil)

(defun add-sort (sort &key nicknames)
  (rewind-question)
  (when nicknames (setp sort 'SHORT-NAME (car nicknames)))
  (dolist (a nicknames) (setp a 'LONG-NAME sort))
  (pushnew sort *sorts*)
  )

;predefined sorts
(add-sort 'episode :nicknames '(ep event))
(add-sort 'number :nicknames '(num))
(add-sort 'propos :nicknames '(prop))
(add-sort 'integer :nicknames '(int))
(add-sort 'real)
(add-sort 'set)
(add-sort 'time)
(add-sort 'string :nicknames '(str))
(add-sort 'operator :nicknames '(op))
;(add-sort 'wff-op) 
(add-sort 'function :nicknames '(func))
(add-sort 'quantifier :nicknames '(quant))


;;;   OPERATORS

(defun add-operator (operator &key op-type result-type arg-types specialists indicators)
  (rewind-question)
  (setp operator 'OPERATOR t) 
  (setp operator 'entity-type 'OPERATOR)
  (setp operator 'OPERATOR-TYPE op-type) 
  ; if op-type, copy its result and arg types here?
  (when result-type (setp operator 'RESULT result-type))
  (when arg-types (setp operator 'ARGS arg-types))
  (when specialists (setp operator 'specialists 
			  (append (getp operator 'specialists) specialists)))
  (when indicators (apply 'add-indicate (append (list operator) indicators)))
  )

(defun add-operator-type (op-type &key result-type arg-types)
  (rewind-question)
  (setp op-type 'RESULT result-type)
  (setp op-type 'ARGS arg-types)
  )

; predefined operator types
(add-operator-type 'wff-op :result-type 'wff :arg-types 'wff)
(add-operator-type 'pred-op :result-type 'pred :arg-types 'pred)
(add-operator-type 'term-op :result-type 'term :arg-types 'term)
(add-operator-type 'quantifier-op :result-type 'quantifier :arg-types 'quantifier)
(add-operator-type 'operator-op :result-type 'operator :arg-types 'operator)
(add-operator-type 'term-forming-op :result-type 'term :arg-types 'pred)

;predefined operators
(add-operator 'nec :op-type 'wff-op)
(add-operator 'poss :op-type 'wff-op)
(add-operator 'prob :op-type 'wff-op)
(add-operator 'almost :op-type 'pred-op)
(add-operator 'sort-of :op-type 'pred-op)
(add-operator 'very :op-type 'pred-op)
(add-operator 'extremely :op-type 'pred-op)
(add-operator 'severely :op-type 'pred-op)
(add-operator 'plural :op-type 'pred-op)
(add-operator 'plur :op-type 'pred-op)
(add-operator 'coll :op-type 'pred-op)

(add-operator 'coll-of :result-type 'pred-op :arg-types 'term)
(add-operator 'member-type-of :result-type 'pred :arg-types 'term)

(add-operator 'set-of-all :op-type 'term-forming-op)
(add-operator 'K :op-type 'term-forming-op)  ; kind
(add-operator 'Ka :op-type 'term-forming-op) ; kind of action
(add-operator 'K1 :op-type 'term-forming-op)
(add-operator 'to :op-type 'term-forming-op)

(add-operator 'that :result-type 'term :arg-types 'wff)
(add-operator 'Ke :result-type 'term :arg-types 'wff)  ; kind of event
(add-operator 'YN-q :result-type 'term :arg-types 'wff)
(add-operator 'whether :result-type 'term :arg-types 'wff)
(add-operator 'answer-to :result-type 'term :arg-types 'wff)

(add-operator 'adv-a :result-type 'pred-op :arg-types 'pred)
(add-operator 'adv-q :result-type 'pred-op :arg-types 'pred)

(add-operator 'adv-s :result-type 'wff-op :arg-types 'pred)

(add-operator 'pres :op-type 'wff-op)
(add-operator 'past :op-type 'wff-op)
(add-operator 'futr :op-type 'wff-op)
(add-operator 'perf :op-type 'wff-op)
(add-operator 'prog :op-type 'wff-op)
(add-operator 'prog :op-type 'wff-op)
(add-operator 'box :op-type 'wff-op)
(add-operator 'perhaps :op-type 'wff-op)

(add-operator 'adv-e :result-type 'wff-op :arg-types 'pred)
(add-operator 'adv-f :result-type 'wff-op :arg-types 'pred)
(add-operator 'adv-p :result-type 'wff-op :arg-types 'pred)
(add-operator 'adv-dest :result-type 'pred-op :arg-types 'pred)

(add-operator 'attr :result-type 'pred-op :arg-types 'pred)
(add-operator 'nn :result-type 'pred-op :arg-types 'pred)
(add-operator 'na :result-type 'pred-op :arg-types 'pred)

(add-operator 'ly :result-type 'pred-op :arg-types 'pred)
(add-operator 'nearly :result-type 'quantifier :arg-types 'quantifier)
(add-operator 'mos :result-type 'pred :arg-types 'pred)  


;;;;    FUNCTIONS  

(defun add-function (fnname &key rel-pred specialists package sort)
  (rewind-question)
  (setp fnname 'FUNCTION t)
  (when sort (setp fnname 'SORT sort))
  (when specialists (setp fnname 'specialists 
			  (append (getp fnname 'specialists) specialists)))
  (when rel-pred (setp fnname 'rel-pred rel-pred))
  (when package (other-function package fnname))
  )

;predefined functions
(add-function 'pair)
(add-function 'set-of)
(add-function 'cardinality-of)
(add-function 'date)
(add-function 'start-of)
(add-function 'end-of)
(add-function 'fst)
(add-function 'rst)
(add-function 'F)


;;;;     PREDICATES

(defun add-predicate (pred &key type part hier parent subnodes hier-type
			   part-type
			   indicators specialists package entry-rtn sort)
  (rewind-question)
  (setp pred 'entity-type 'PRED)
  (setp pred 'PREDICATE t)
  (when indicators (apply 'add-indicate (append (list pred) indicators)))
  (when package 
	(other-predicate package pred)
	(when entry-rtn (other-entry-rtn package pred entry-rtn)) )
  (when specialists (setp pred 'specialists 
			  (append (getp pred 'specialists) specialists)))
  (when (or type part hier parent subnodes hier-type)
	; if hier already exists - type,part info not nec, but check it
	(if (and (null hier) parent) (setq hier (car (getp parent 'hier))))
	(when (getp hier 'root)
	      (cond (type 
		     (if (or (getp hier 'pred-hier) 
			     (getp hier 'part-hier)) 
			 (return-from add-predicate
				      (print-line "*** ERROR: " hier 
				     " is not a type hierarchy"))))
		    (part
		     (if (not (getp hier 'part-hier))
			 (return-from add-predicate 
				      (print-line "*** ERROR: " hier 
				     " is not a part hierarchy"))))
		    (part-type
		     (if (not (getp hier 'part-type-hier))
			 (return-from add-predicate 
				      (print-line "*** ERROR: " hier 
				     " is not a part type hierarchy"))))
		    (t (cond ((getp hier 'part-hier) (setq part t))
			     ((getp hier 'part-type-hier) (setq part-type t))
			     ((getp hier 'pred-hier) nil)
			     (t (setq type t))))
		    ))
	; if hier doesn't exist, type indicates type hier, part indicates
	; part hier, and neither indicates pred hier
	; if type or part specified and no hier name, invent one same as 
	; pred (if poss)
	(when (and (or part type) (null hier)) 
	      (cond (part (setq hier 'part)
			  (setq parent 'part))
		    (t (setq hier pred))))
	(if hier 
	    (cond (part-type (set-hier-type hier 'part-type-hier))
		  ((and (not part) (not type)) (set-hier-type hier 'pred-hier)))
	    )
	(if hier-type (set-hier-type hier hier-type))
	; then call add-hier or add-part-hier as necessary
	(unless (or parent subnodes) (add-hier hier pred))
	(when parent
	      (if part
		  (add-part-hier hier parent pred)
		  (add-hier hier parent pred)))
	(when subnodes
	      (if part
		  (apply 'add-part-hier (append (list hier pred) subnodes))
		  (apply 'add-hier (append (list hier pred) subnodes))
		  ))
	(when sort (add-sort-to-hier pred sort))
	)
  )

;predefined predicates
(add-predicate 'entity :type t :hier 'entity :hier-type 'overlap)

(add-predicate 'meta-entity :type t :hier 'meta-entity)
(add-predicate 'wff :type t :sort 'wff :hier 'meta-entity :parent 'meta-entity)
(add-predicate 'operator :type t :sort 'operator :hier 'meta-entity 
		     :parent 'meta-entity)
;(add-predicate 'wff-op :type t :sort 'wff-op :hier 'meta-entity 
;		     :parent 'operator)
(add-predicate 'function :type t :sort 'function :hier 'meta-entity 
		     :parent 'meta-entity)
(add-predicate 'quantifier :type t :sort 'quantifier :hier 'meta-entity 
		     :parent 'meta-entity)
(add-predicate 'pred :type t :sort 'pred :hier 'meta-entity 
		     :parent 'meta-entity)
(add-predicate 'term :type t :sort 'term :hier 'meta-entity 
                     :parent 'meta-entity)
(add-predicate 'e-term :type t :sort 'e-term :hier 'meta-entity :parent 'term)
(add-predicate 'numeral :type t :sort 'numeral :hier 'meta-entity :parent 'term)

(add-predicate 'sorted-entity :type t :hier 'entity :parent 'entity)
(add-hier 'term 'term 'sorted-entity)
(add-predicate 'episode :type t :sort 'episode :hier 'entity :parent 'sorted-entity)
(add-predicate 'number :type t :sort 'number :hier 'entity :parent 'sorted-entity)
(add-predicate 'integer :type t :sort 'integer :hier 'entity :parent 'number)
(add-predicate 'real :type t :sort 'real :hier 'entity :parent 'number)
(add-predicate 'set :type t :sort 'set :hier 'entity :parent 'sorted-entity)
(add-predicate 'time :type t :sort 'time :hier 'entity :parent 'episode)
(add-predicate 'propos :type t :sort 'propos :hier 'entity :parent 'sorted-entity)
(add-predicate 'string :type t :sort 'string :hier 'entity :parent 'sorted-entity)
(add-predicate 's-list :type t :sort 's-list :hier 'entity :parent 'sorted-entity)

(add-predicate 'part :part-type t :hier 'part :hier-type 'overlap)

(add-predicate '**)
(add-predicate '*)
(add-predicate '@)
(add-predicate '_)
(add-predicate 'because)

;;;; QUANTIFIERS

(defun add-quantifier (quant &key prob negation distributive)
  (rewind-question)
  (setp quant 'QUANTIFIER t) 
  (setp quant 'entity-type 'QUANTIFIER)
  (when prob (quantifier-prob quant prob))
  (when negation (setp quant 'negation negation))
  (when distributive (setp quant 'distributive distributive))
  )

(add-quantifier 'A :prob 1)
(add-quantifier 'E :prob 1)
(add-quantifier 'E! :prob 1)
(add-quantifier 'THE :prob 1)
(add-quantifier 'WH :prob 1)
(add-quantifier 'most :prob 0.8)
(add-quantifier 'many :prob 0.6)
(add-quantifier 'some :prob 0.4)
(add-quantifier 'few :prob 0.2)
(add-quantifier 'none :prob 0.0)



