;; Last change 29 March 2005 - FM finally solved (probably) the bug
;; found by Alexander on his email: Date: Tue, 01 Mar 2005 18:44:25
;; -0500 From: Alexander S. Yeh <asy@mitre.org> To: EPILOG
;; <epilog@cs.rochester.edu> Cc: Samuel L. Bayer <sam@mitre.org>
;; Subject: Re: Combining probabilities?).  The problem was that the
;; WFF is a symbol loaded in the package EPILOG but here we are in the
;; EPILIB package so every time we do a (getp wff 'some-symbol)
;; some-symbol is actually epilib::some-symbol so it doesn't match any
;; property of wff. Solved using (intern 'some-symbol)... by far a
;; good solution. This modification is on the aux arguments of the
;; function recalc-prob. FM.

;;       Last change    August 1993
(in-package epilib)

;
; Probability calculation and manipulation routines
;

(traceable 'prob)

(defvar *prob-already* nil)

(defun gimme-prob (w &aux
		     (p (getp w 'PROB)))
  (cond ((member w *prob-already*)
	 (list 1))
	(t (push w *prob-already*)
	   (if (listp p)
	       (apply 'append 
		      (mapcar #'(lambda (a) 
					(if (symbolp a)
					    (cond ((member a *prob-already*)
						   nil)
						  (t (gimme-prob a)))
					    (list a)))
			      p))
	       (list p)) ) )
  )

(defun calc-prob-list (plist)
  (setq *prob-already* nil)
  (if (listp plist)
      (apply '* (apply 'append 
	  (mapcar #'(lambda (p) (if (numberp p) (list p) (gimme-prob p)))
		  plist)) )
      plist)
  )

(defun calc-prob (wff-name) (calc-prob-list (getp wff-name 'PROB)))

(defun combine-probs (p1 p2)
  (let ((res (cond ((null p1) p2)
		   ((null p2) p1)
		   ((numberp p1)
		    (cond ((numberp p2) (* p1 p2))
			  ((numberp (car p2)) (append (list (* p1 (car p2))) (cdr p2)))
			  (t (append (list p1) p2))))
		   ((numberp (car p1))
		    (cond ((numberp p2) (append (list (* (car p1) p2)) (cdr p1)))
			  ((numberp (car p2)) (append (list (* (car p1) (car p2)))
						      (union (cdr p1) (cdr p2))))
			  (t (append (list (car p1)) (union (cdr p1) p2)))))
		   (t (cond ((numberp p2) (append (list p2) p1))
			    ((numberp (car p2)) 
			     (append (list (car p2)) (union (cdr p2) p1)))
			    (t (union p1 p2))))
		   )
	     )
	)
    (when (traced-p 'prob)
	(format t "Combine-probs: p1=~a p2=~a result=~a~%" p1 p2 res)
      )
    res
    )
  )

; get-prob returns the probability to use in ordering the clauses within
; a classification. For numeric conditionals, the predicate (number) is used

(defun get-prob (wff-name &aux pred
			  (ttype (getp wff-name 'wff-type)))
  (cond ((and (equal ttype 'quantified)
	      (not (getp wff-name 'conditional)))
	 (getp wff-name 'C-PROB))
	((getp wff-name 'conditional)
	 (if (and (setq pred (getp wff-name 'pred))
		  (numberp pred))
	     pred
	     (getp wff-name 'C-PROB)))
	(t (getp wff-name 'c-prob)) )
  )

(defun combine-contradict-probs (prob1 prob2 &aux 
				       (p1 (calc-prob-list prob1))
				       (p2 (calc-prob-list prob2))
				       (prod (* p1 p2)))
  (/ (- p1 prod) (- 1 prod)) 
  )

(defun combine-support-probs (problist &aux 
				       (plist (mapcar 'calc-prob-list problist))
				       prob) 
  (do ((p (car plist) (car plist)))
      ((null p))
      (setq plist (cdr plist))
      (if (null prob)
	  (setq prob p)
	  (setq prob (- 1 (* (- 1 prob) (- 1 p))))
	  )
      )
  prob)

(defvar *recalced-wffs* nil)

(defun recalc-probs (wff)
  (setq *recalced-wffs* nil)
  (recalc-prob wff)
  )

(defun recalc-prob (wff &optional (phi (getp wff 'print)) &aux
					(prob (getp wff (intern 'prob)))
							  (c-prob (getp wff (intern 'c-prob)))
							  (contradict (getp wff (intern 'contradict)))
							  (support (getp wff (intern 'support)))
							  sup con newprob)
  (unless (member wff *recalced-wffs*)
    (push wff *recalced-wffs*)
    (if support (setq sup (combine-support-probs support)))
    (if contradict (setq con (combine-support-probs contradict)))
    (setq newprob
      (if sup
	  (if con
	      (combine-contradict-probs sup con)
	    sup)
	(if con
	    con
	  (calc-prob-list prob))))
    (when (if c-prob (not (= c-prob newprob)) newprob)
      (if (equal prob c-prob) (setp wff 'prob newprob))
      (setp wff 'c-prob newprob)
      (when (traced-p 'forward)
	(print-line " Recalculated probability of " wff ": " phi 
		    " to " newprob))
      (update-descendant-probs wff))
    newprob))

(defun update-descendant-probs (wff)
  (mapcar 'recalc-prob (getp wff 'descendants))
  )

; ---------------------------------------------------------------
;  New combined probability calculations using method by 
;   Bulent Murtezaoglu

(defun calc-combined-prob (lp1 up1 lp2 up2 &optional (coop t)
			       neg-lp1 neg-up1 
			       neg-lp2 neg-up2)
  ; coop is t if the formulas support each other, nil if they contradict
  (when (null up1) (setq up1 1.0))
  (when (null up2) (setq up2 1.0))
  (when (null neg-lp1)
	(setq neg-lp1 (- 1.0 up1))
	(setq neg-up1 (- 1.0 lp1)) )
  (when (null neg-lp2)
	(setq neg-lp2 (- 1.0 up2))
	(setq neg-up2 (- 1.0 lp2)) )
  (when (null neg-up1) (setq neg-up1 1.0))
  (when (null neg-up2) (setq neg-up2 1.0))
  ; calculate new bounds on the SAME formula - i.e. if they contradict, use
  ; the bounds for the negated version of one of them - then we are combining
  ; bounds for the positive versions, or the negative versions, but not mixed.
  (if coop
      (calc-prob-bounds lp1 up1 lp2 up2)
      (if (> lp1 lp2)
	  (calc-prob-bounds lp1 up1 neg-lp2 neg-up2)
	  (calc-prob-bounds neg-lp1 neg-up1 lp2 up2)
	  ))
  )

(defun calc-prob-bounds (lp1 up1 lp2 up2 &aux newlp newup)
  ; problems with 0 as one of the bounds
  (cond (nil ;(or (= lp1 0) (= lp2 0))
	 (setq newlp (calc-new (- 1.0 up1) (- 1.0 up2)))
	 (setq newup (calc-new (- 1.0 lp1) (- 1.0 lp2)))
	 (list (- 1.0 newup) (- 1.0 newlp)))
	(t (setq newlp (calc-new lp1 lp2))
	   (setq newup (calc-new up1 up2))
	   (list newlp newup))
	)
  )

(defun calc-new (p1 p2)
  ; what to do if p1 or p2 is 0 ??
  (cond ((and (> p1 0) (> p2 0))
	 (/ 1.0 (+ 1.0 (sqrt (abs (- (/ (+ 1.0 (* -1 p1) (* -1 p2) (* 2 p1 p2)) 
			       (* p1 p2)) 1.0)))) ))
	(t (/ (+ p1 p2) 2)))
  )

