;       Last change    August 1993
(in-package lib)


(defvar *display-items* nil "Items which may be displayed using display")
(defvar *display-default* nil)

(defun set-display-function (name func-name &optional description)
  (setp name 'display-function func-name)
  (if description (setf (documentation name 'setf) description))
  (pushnew name *display-items*)
  )

; print-display prints all display items or a selection of them
; given by the user. The item's name, and its description are printed.
 
(defun print-display (&rest items)
  (if items 
      (setq items (intersection items *display-items*))
      (setq items *display-items*))
  (dolist (item items)
	  (print-it *nl* " "  (symbol-name item) ": " (documentation item 'setf)
		    ))
  )

(defun pr-display (brief &rest args)
  (declare (ignore brief))
  (apply 'print-display args))

(set-display-function 'help 'pr-display "Shows display options available")
(set-display-function '? 'pr-display "Shows display options available")
(set-display-function 'options 'pr-display "Shows display options available")
(set-display-function 'display 'pr-display "Shows display options available")
(set-display-function 'trace 'pr-display "Prints trace values available")
(set-display-function 'tweak 'pr-display "Prints tweakable parameters")
(set-display-function 'traceable 'pr-display "Prints trace values available")
(set-display-function 'tweakable 'pr-display "Prints tweakable parameters")

(defun display (&rest original-args &aux cmd-args (brief t) func-name key
		      tweak display trace d)
    (setq cmd-args
        (delete nil (mapcar #'(lambda (arg)
                                (case arg 
                                    ((-brief -b) (setq brief t) nil)
                                    ((-full -f) (setq brief nil))
                                    ((-key -k) (setq key t) nil)
				    ((tweak tweakable) (setq tweak t) nil)
				    ((display) (setq display t) nil)
				    ((trace traceable) (setq trace t) nil)
                                    (otherwise arg)))
                           original-args )))
    (cond (key
	   (let (alist) 
		(when (null cmd-args)
		      (when tweak
			    (setq cmd-args (append cmd-args (list 'tweak)))
			    (setq tweak nil))
		      (when trace
			    (setq cmd-args (append cmd-args (list 'trace)))
			    (setq trace nil))
		      (when display
			    (setq cmd-args (append cmd-args (list 'display)))
			    (setq display nil))
		      )
		(dolist (a cmd-args) 
			(setq alist (append alist (apropos-list a))))
		(when alist
		      (setq alist (remove-duplicates alist))
		      (if (or display (and (null trace) (null tweak)))
			  (apply 'print-display alist))
		      (if tweak (apply 'pr-tweak (cons brief alist)))
		      (if trace (apply 'print-traceables alist))
		      )
		))
	  (tweak (apply 'pr-tweak (cons brief cmd-args)))
	  (trace (apply 'print-traceables cmd-args))
	  (display (apply 'print-display cmd-args))
	  ((not cmd-args) (print-display))
	  ((setq func-name (if (symbolp (car cmd-args)) 
			       (getp (car cmd-args) 'display-function)))
	   (if (or (not (fboundp func-name))
		   (macro-function func-name)
		   (special-operator-p func-name))
	       (print-error "The Lisp function definition for DISPLAY "
			    (car cmd-args) " (the function " func-name
						  ") is missing.")
	       (apply (symbol-function func-name) (cons brief (cdr cmd-args))))
	   )
	  ((subsetp cmd-args (append *tweakable-items* *traceable-items*))
	   (dolist (c cmd-args)
		   (cond ((member c *tweakable-items*)
			  (print-it *nl* c "    Tweakable parameter ")
			  (when (setq d (documentation c 'setf))
				(print-it *nl* "    " d))
			  (print-it *nl* "    Default value: " (getp c 'default))
			  (print-it *nl* "    Current value: " (eval c)))
			 ((member c *traceable-items*)
			  (print-it *nl* c "    Trace value ")
			  (when (setq d (documentation c 'setf))
				(print-it *nl* "    " d))
			  (when (getp c 'trace-values) 
				(print-it *nl* "    Starts tracing of: " 
					    (getp c 'trace-values)))
			  (if (if (getp c 'trace-values)
				  (null (set-difference (getp c 'trace-values)
							*traced-items*))
				  (member c *traced-items*))
			      (print-it *nl* "    Currently being traced")
			      (print-it *nl* "    Not currently being traced")))
			 )) )
	  (*display-default* 
	   (apply #'display (cons *display-default* original-args)))
	  ((and (null (cdr original-args)) (symbolp (car original-args)))
	   (print-it *nl* (car original-args) ": " 
		     (symbol-plist original-args)))
	  (t (print-it *nl* original-args))
	  )
  nil)


