;       Last change    August 1993
(in-package part)

; the following routines are used to determine whether two literals are 
; resolvable or factorable


; classify-lits classifies the given literals, and returns a list
; of substitution orders and which literal to enter.
; Each item in the list is of the form (orderlist . number of lit to enter)
; Each item in orderlist is of the form 
;     (element number from 1st lit . element number from second)

(defun classify-lits (neg-p1 pred1 neg-p2 pred2) 
  ; each literal has only 2 args which can be unified
  ; only two possible unifications - same direction and reversed
  (declare (ignore neg-p1 pred1 neg-p2 pred2))
 (list '((1 . 1)(2 . 2))) ;'((1 . 2)(2 . 1))
  )


; incompatible-add checks to make sure that the residues are not the same as the
; original literals (without substitutions). If so, a resolving item is
; added, otherwise, nothing is done.

(defun incompatible-add (lit-1 lit-2 res1 res2 eval1 eval2 subs1 subs2
			       &optional (incompatible t)
			       &aux rr)
  (declare (ignore eval1 eval2))
  (unless (or (member lit-1 res1)
	      (member lit-1 res2)
	      (member lit-2 res1)
	      (member lit-2 res2))
          (when (traced-p 'epilog::part-test)
		(print-line "    Part Specialist: Found " (print-info lit-1 )
		     (if incompatible " incompatible" " compatible") " with "
		     (print-info lit-2) )
		(print-line "       with residues: "
			    (if (setq rr (append res1 res2)) rr "null"))
		(print-line "       with substitutions: " 
			    (print-subs (append subs1 subs2))))
	  (list 
	   (make-comparison-info :lit1 lit-1 :lit2 lit-2
			     :subs1 subs1 :subs2 subs2 
			     :residue1 res1
			     :residue2 res2
			     ))
	  )
  )


; create-residue creates a literal of the given neg pred ard args

(defun create-residue (neg pred &rest args &aux
			       (reln (find-symbol (symbol-name pred) 'epilog)))
  (if (or (null pred)
	  (null args)
	  (eq (car args) (cadr args)))
      t  
      (make-prop neg :pred reln :args args)
      )  
  )

; test-residue does the necessary substitutions, and then resolves or
; factors the resulting literals.

(defun test-residue (incompatible net lit-1 lit-2 ordlist
			     neg-p1 reln1 arglist1
			     neg-p2 reln2 arglist2 effort &aux
			     (newargs1 arglist1)
			     (newargs2 arglist2)
			     (eval1 'unknown)
			     (eval2 'unknown)
			     newlit1 newlit2
	                     ulist subs1 subs2)
  (unless (null ordlist)
	  ; unify the two literals according to specification given
          (setq ulist (unify-lists arglist1 arglist2 ordlist nil))

	  ; make substitutions in the two literals
          (unless (null ulist)
                  (setq subs1 (car ulist))
	          (setq subs2 (cdr ulist))
	          (setq newlit1 (normalize (substitute-prop lit-1 subs1) '+ nil nil))
	          (setq newlit2 (normalize (substitute-prop lit-2 subs2) '+ nil nil))
		  ))

  (unless (or (null newlit1) (null newlit2))
          
	  ; evaluate the arguments 
	  (setq newargs1 (spec-eval-args newlit1))
	  (setq newargs2 (spec-eval-args newlit2))

	  ; evaluate the literals
	  (setq eval1 (eval-with-neg net neg-p1 reln1 newargs1 effort))
	  (setq eval2 (eval-with-neg net neg-p2 reln2 newargs2 effort))

	  ; make sure substitution list complete
          (when (equal eval1 'unknown)
                (setq subs1 (union subs1 
		                (arg-sub-list newargs1 
						   (getp newlit1 'arguments) 1)))
	        (setq newlit1 (normalize (substitute-prop lit-1 subs1) '+ nil nil))
	        )
          (when (equal eval2 'unknown)
                (setq subs2 (union subs2 
		              (arg-sub-list newargs2 
						 (getp newlit2 'arguments) 2)))
	        (setq newlit2 (normalize (substitute-prop lit-2 subs2) '+ nil nil))
	        )

	  ; then use evaluations if possible

              (incompatible-with-eval lit-1 lit-2 eval1 eval2 subs1 subs2 
		                        (traced-p 'epilog::part-test)
		                        "    Part Specialist:"
					incompatible)
	  )
  )

; part-test returns t if both predicates are part-of predicates, nil
; otherwise 

(defun part-test (neg-p1 reln1 neg-p2 reln2)
  (declare (ignore neg-p1 neg-p2))
  (and (member reln1 '(part-of))
       (member reln2 '(part-of)))
  )


; unify-test-subpart unifies the arguments, performs substitutions and
; calls test-subpart to do the actual resolving or factoring attempt

(defun unify-test-subpart (incompatible net lit-1 neg-p1 reln1 arglist1 
				  lit-2 neg-p2 reln2 arglist2 effort &aux
				  newp11 newp21
			    (p11 (car arglist1))
			    (p12 (cadr arglist1))
			    (p21 (car arglist2))
			    (p22 (cadr arglist2))
			    subs)
  (setq subs (compare-arglists (list p11) (list p21) :cleanup t))
  (unless (null subs)
	  (setq newp11 (substitute-term  p11 (car subs) (cdr subs)))
	  (setq newp21 (substitute-term  p21 (car subs) (cdr subs)))
	  (test-subpart incompatible net lit-1 neg-p1 newp11 reln1 p12
			   lit-2 neg-p2 newp21 reln2 p22 
			   (car subs) (cdr subs) effort)
	  )
  )


; test-subpart determines whether two literals are resolvable or
; factorable and the resulting residues or factors, and adds these
; actions to the agenda  

(defun test-subpart (incompatible net lit-1 neg-p1 part11 reln1 part12
				lit-2 neg-p2 part21 reln2 part22 
				subs1 subs2 effort)
  (declare (ignore incompatible net lit-1 neg-p1 part11 reln1 part12
				lit-2 neg-p2 part21 reln2 part22 
				subs1 subs2 effort))
  )


; incompatible-lits tries to determine whether lit1 and lit2 are resolvable
; it tries all possible substitutions to find out

(defun incompatible-lits (netname lit1 lit2 neg-p1 reln1 arglist1 
		   	   neg-p2 reln2 arglist2 effort 
			   &optional (incompatible t) &aux
		           (net (determine-net netname))
			   (all-constants (not (or
					   (non-constant-arg arglist1)
					   (non-constant-arg arglist2))))
			   )

  ; see if arguments unifiable, and in what order

  (when (traced-p 'epilog::part-test)
        (print-line "    Part Specialist: Trying to determine "
		  (if incompatible "incompatibility" "compatibility") " of " 
		  (print-info lit1) " and " (print-info lit2))
	(if (not (equal (part-subnet-identifier net) '/))     
	    (print-it "    in subnet " (part-subnet-identifier net)))
      )

  (if all-constants
      (let ((eval1 (eval-with-neg net neg-p1 reln1 arglist1 effort))
	    (eval2 (eval-with-neg net neg-p2 reln2 arglist2 effort)))
	   ; try evaluations first
	   ;unless
	    (cond ((incompatible-with-eval lit1 lit2 eval1 eval2 nil nil
				      (traced-p 'epilog::part-test)
				      "    Part Specialist:" incompatible))
		  ((and (equal (car arglist1) (car arglist2))
			(part-test neg-p1 reln1 neg-p2 reln2))
		   (test-subpart incompatible net lit1 neg-p1
				 (car arglist1) reln1 (cadr arglist1) lit2
				 neg-p2 (car arglist2) reln2 (cadr arglist2)
				 nil nil effort))
		  )
	   )

      ; if not constants
      (if (part-test neg-p1 reln1 neg-p2 reln2)
	  (unify-test-subpart incompatible net lit1 neg-p1 reln1 arglist1 
				 lit2 neg-p2 reln2 arglist2 effort)

	  ; otherwise try entering one and testing the other
	  (if (and (type-pred-p reln1) (type-pred-p reln2))
      (let* ((ord-entry (classify-lits neg-p1 reln1 neg-p2 reln2)))
	    (if (null ord-entry)
	        nil
		(let (r)
		     (mapcar #'(lambda (o) (setq r (append r
			 (test-residue incompatible net lit1 lit2 o neg-p1 reln1
				       arglist1 neg-p2 reln2 arglist2 effort))))
				      ord-entry)
		     r)
		)
	    )))
      )
  )

; compatible-lits tries to determine whether lit1 and lit2 are factorable.
; it tries all possible substitutions to do this.

(defun compatible-lits (netname lit1 lit2 neg-p1 pred1 arglist1 
		   	   neg-p2 pred2 arglist2 effort )
  (incompatible-lits netname lit1 lit2 neg-p1 pred1 arglist1 neg-p2 pred2
		     arglist2 effort nil)
  )
