;       Last change    August 1993
(in-package response)

; Filtering removes clauses which are obvious to the user so that they are
; not included in the answer - note that the set-of-support clauses have
; already been removed by this point.

(tweakable '*filter-threshold* 100 ;50 
  "Threshold of known likelihood beyond which to ignore clauses in answer"
  '(float * *))
(tweakable '*max-response-complexity* 55 ; 50 
  "Maximum complexity of wff to try to say - if nil, no maximum")
(tweakable '*prompt-if-too-complex* nil ;50 
  "Indicates whether or not to ask user before ignoring a complex wff in response"
  'boolean)


; filter-support is used to remove the set-of-support clauses retireved to 
; build a noun phrase description

(defun filter-support (plist)
  (remove-if 'filter-check plist)
  )

(defun filter-check (item &optional neg)
  (or (getp item 'PROOF) (getp item 'DISPROOF)
      (if neg (getp item 'negated)))
  )

(defun filter-complex (plist)
  (remove-if 'complex-check plist))

(defconstant responses '(Y N))

(defun complex-check (item)
  (cond ((null *max-response-complexity*) nil)
	((null (getp item 'complexity)) nil) ; a specialist evaluation
	((> (getp item 'complexity) *max-response-complexity*)
	 (cond (*prompt-if-too-complex*
		(print-line "*** " item ": " (print-info item) 
			    *nl* "       is too complex to print")
		(if (y-or-n-p "Do you want me to try anyway?") nil t) )
	       (t (if (traced-p 'filtration)
		      (print-line "  Response: Ignoring " item ": " (print-info item)
			     *nl* "      because its complexity is too high ("
				  (getp item 'complexity) ")"))
		  t)))
	(t nil))
  )

; default-known determines the default likelihood of a clause being known
; simple type predications and specialist evaluations have the highest
; likelihoods - 90
; next highest are clauses consisting of all variables,
; then clauses consisting of mixed variables and constants
; and lowest are clauses consisting of all constants
; perhaps someday include interst as another factor?

(defun default-known (formula &aux
		     (variables (list-length (getp formula 'types)))
		     (constants (list-length (getp formula 'about)))
		     pred)
  (cond ((getp formula 'meaning-postulate) 100)
	
	((and (= constants 0) (> variables 0)) ; all variables
	 80)  ; general knowledge

	((and (= variables 0) (> constants 0)) ; all constants
	 ; if simple type predication, high likelihood
	 ; we can always look it up again if we need it
	 (if (type-pred-p (setq pred (getp formula 'pred)) )
	     (if (equal pred (getp pred 'sort)) 101 90) 40) )

	(t 50))   ; mixed variables and constants, may be story or general info
  )

(set-display-function 'filter 'show-filter-threshold "Shows the current filter threshold")

(defun show-filter-threshold (&rest args)
  (declare (ignore args))
  (print-line "The current filter threshold is " *filter-threshold*)
  )
(defun filter-< (wff1 wff2 &aux p1 p2
		   (kl1 (getp wff1 'KNOWN-LIKELIHOOD))
		   (kl2 (getp wff2 'KNOWN-LIKELIHOOD))
		   )
  (cond ((< kl1 kl2))
	((= kl1 kl2)
	 (cond ((and (getp wff1 'negated) (not (getp wff2 'negated))))
	       ((and (setq p1 (getp wff1 'pred))
		     (setq p2 (getp wff2 'pred))
		     (or (and (member p1 '(* **) :test 'equal)
			      (equal p2 '@))
			 (and (equal p1 '**)
			      (equal p2 '*)))
		     ))
	       ))
	)
  )

; filter checks each clause in the list of clauses to see if it has a
; higher likelihood than the threshold. If so, it is not included in the
; list returned, unless it is the only clause, in which case it stays.

(defun filter-response (givenclauses &aux
		       (clauses (filter-support givenclauses))
		      sortclauses useclauses lastw known)
  ; returns filtered set of clauses
  (setq clauses (filter-complex clauses))
  (dolist (c clauses)
	  (setq known (getp c 'KNOWN-LIKELIHOOD))
	  (unless known
		  (setq known (default-known c))
		  (if (traced-p 'filtration-details)
		      (print-line 
		       "  Response: Estimating likelihood of " 
		       c ": " (getp c 'print) " being known as " known))
		  (setp c 'KNOWN-LIKELIHOOD known)
		  ) )
  ; Now sort clauses according to likelihood, and remove all above the
  ; threshold, unless the last clause
  ; sorted list goes from lowest to highest likelihood
  (setq sortclauses (reverse (sort clauses 'filter-< )))
  (setq lastw (last1 sortclauses))
  (if (> (list-length sortclauses) 1)
      (dolist (s sortclauses)
	      (if (and (> (getp s 'KNOWN-LIKELIHOOD) 
			  *filter-threshold*)
		       (if (equal s lastw) (not (null useclauses)) t) )
		  (if (traced-p 'filtration)
		      (print-line 
		       "  Response: Filtering out clause " s ": "
		       (getp s 'PRINT) *nl* 
		       "           with likelihood of being known "
		       (getp s 'KNOWN-LIKELIHOOD)))
		  (setq useclauses (append useclauses (list s)))
		  ) )
      (setq useclauses sortclauses))
  useclauses)

