;       Last change    August 1993
(in-package user)
(eval-when (compile load eval)   
(provide 'response-pkg)
(require 'lib "epilog:lib;lib")
(require 'epilib "epilog:epilib;epilib")
(defpackage response-pkg (:nicknames response)
  (:use lisp user #+lucid lucid-common-lisp #-lucid common-lisp #+excl excl lib epilib)
  (:export
	  respond
	  add-word add-lex 
	  do-say say-it say-them

          *filter-threshold* *default-lex* *default-trans* *basic-nouns* 
	  *fragment-routines* *response-warn* *active-topics*
	  *prompt-if-too-complex* *max-response-complexity*
	  *fragment-groups* *fragment-types*

	  filtration organization verbalization response 
	  fragment retrieval
	  combine combine-details response-all filtration-details
	  trans-details

	  filter lex trans

	  ans answer more more-details details less-details less 

	  I O S D ED ES ING H ; A E U W X Y Z already exported from epilib
	  NP VP AP ADVP ADVS PP RelC
          proper-noun adj pronoun required
          objects particles name parts 
          number plural singular
          pre conj if then only but  ;and or
          prep of full
          determiner all every that what who
          adv usually frequently sometimes occasionally never always rarely
          noun part thing person human likely start end location instance
	  either both removed as
          verb be do have by will member fact one group feminine masculine mass
          aspect passive prespart infinitive no-passive progressive perfect
          tense present future pastpart perf-prog mood indicative subjunctive
	  imperative

	  before after while overlaps overlapped-by because when

	  make-adverb make-passive force-adverb
	  ))
(unless (find-package 'response) (make-package 'response)))
(in-package response)

(traceable 'filtration "Filtration of set of support clauses")
(traceable 'filtration-details "Shows estimation of likelihood for filtration")
(traceable 'organization "Organization of wffs for response")
(traceable 'verbalization "Verbalization of wffs during response generation")
(traceable 'response "Response process")
(traceable 'trans-details "Details of translation")
(traceable 'fragment "Fragment details in response generation")
(traceable 'retrieval "Retrieval of additional info during response generation")
(traceable 'combine "Combination of fragments during response generation")
(traceable 'combine-details "Details of combination during response generation")
(traceable 'response-all "Everything to do with response generation"
  '(filtration organization verbalization response fragment retrieval combine 
    combine-details))
(traceable 'response-int "Interesting stuff during response generation"
  '(filtration organization verbalization response fragment retrieval ))
(traceable 'response-min "Minimum stuff to see how response generation works"
  '(filtration verbalization response retrieval))

(load "epilog:response;defs")
(load "epilog:response;lex")
(load "epilog:response;filter")
(load "epilog:response;organize")
(load "epilog:response;general")
(load "epilog:response;combine")
(load "epilog:response;fragment")
(load "epilog:response;build")
(load "epilog:response;fill")
(load "epilog:response;print")
(load "epilog:response;trans")
(load "epilog:response;verbalize")

(defvar *last-answer* nil "Last set of wffs used in answer")
(defvar *old-answers* nil "Last sets of wffs used in answer")

(defconstant *response-properties* 
  '(member group mass masculine 
    feminine)
  "List of lexical properties recognized by the response generator")

(defun determine-answer-clauses (answer detail)
  (cond ((null *last-answer*) 
	 (setq *old-answers* nil)
	 (setq *last-answer* (cadr answer)))
	((member detail '(more-details more details))
	 (push *last-answer* *old-answers*)
	 (setq *last-answer*
	     (mapcan #'(lambda (a) (getp a 'PARENTS)) *last-answer*)))
	((member detail '(less less-details))
	 (if *old-answers* (setq *last-answer* (pop *old-answers*))))
	(t (setq *old-answers* nil)
	   (setq *last-answer* (cadr answer)))
	))

; respond is the routine that takes the answer and the clauses used to get
; that answer and genrates an English answer out of them.

(defun respond (answer &optional detail &aux
		       clauses
		 fcl ocl)
  ; answer of the form (yes clauses prob)
  (when (listp answer)
	(if (listp (caadr answer)) 
	    (setq answer (list (car answer) (caadr answer) (caddr answer))))
  (setq clauses (determine-answer-clauses answer detail)))

  (if (and (listp answer)
	   (member (car answer) '(YES NO) :test #'equal))
      (let ()
	   (setq fcl (filter-response clauses))
           (when (and fcl (traced-p 'response))
                 (print-line "  Response: Clauses to use in response:")
	         (mapcar 
	           #'(lambda (c) (print-line "  " (getp c 'PRINT)))
		  ocl))
           (setq ocl (organize fcl))
           )
      (setq ocl nil))
  (verbalize (if (listp answer) (car answer) nil) ocl)
  )

(defun say-them (clist &aux
		       (fcl (mapcar 'filter-response 
				  (if (listp clist) clist (list (list clist)))))
		       answer ocl)
  (when (and fcl (traced-p 'response))
        (print-line "  Response: Clauses to say:")
	(dolist (cl fcl)
		(mapcar #'(lambda (c) (print-line "  " 
						  (getp c 'PRINT)))
			cl))
	)
  (setq ocl (mapcan 'organize fcl))
  (verbalize answer ocl)
  )

; say-it takes a number of clauses and tries to say them in English. 
; It is similar to respond, except that no filtering is done, and there is no
; answer.

(defun say-it (clauses &aux
		       (fcl (filter-response (if (listp clauses) clauses (list clauses))))
		       answer ocl)
  (when (and fcl (traced-p 'response))
        (print-line "  Response: Clauses to say:")
	(mapcar #'(lambda (c) (print-line "  " (getp c 'PRINT)))
		fcl))
  (setq ocl (organize fcl))
  (verbalize answer ocl)
  )

;;; do-say says a group of wffs retrieved or the answer. Two options are allowed:
;  do-say ans (give the English response for the last question answered)
; and say retrieve parameters, which retrieves the clauses specified by
; parametsrs (using the retrieve command) and says them

(defun do-say (&rest args &aux savep
		     (type (car args))
		     (useargs (cdr args)))
  (case type
	; repeat last answer
	((answer ans)
	 (if *wh-pattern*
	     (mapcar 'say-it (mapcar 'cadr *solutions*))
	     (respond (car *solutions*))) )

	((more-details more details)
	 (respond (car *solutions*) type))

	((less-details less)
	 (respond (car *solutions*) type))

	; say an arbitrary group of clauses
	((retrieve ret)
	 (let* ((args (if (cdr useargs) useargs
			  (list (car useargs) 'tp.topics t t)))
		(cl (apply #'retrieve args)))
	     ; (say-it cl)
	     (say-them (mapcar 'list cl))
	      ) 
	 )

	(otherwise
	 (cond ((eval (append (list 'and) (mapcar #'(lambda (a) 
				   (if (and (symbolp a) (getp a 'wff-type)) t nil)) args)))
		(setq savep *prompt-if-too-complex*)
		(setq *prompt-if-too-complex* t)
		(apply 'say-it args)
		(setq *prompt-if-too-complex* savep)
		nil)
	       (t (print-line
	      "Invalid arguments to do-say - must indicate answer or retrieve"))
	       ))
	)
  )

