;       Last change    August 1993
(in-package epilog)

;; Access actions
;;   Partial accesses are allowd. When a set of wffs is retrieved using
;;   a given classification, they are compared one by one against the
;;   clause given. When a match is found, the remaining wffs are kept
;;   on the access agenda item, and the item is placed back on the agenda.
;;   This prevents doing comparisons on a whole group of wffs when any
;;   one of the group would suffice, yet the other wffs are still available
;;   for comparison if that match doesn't lead to an answer.

(tweakable '*max-wffs* 10 
  "Maximum number of wffs to test for an access action" '(integer 0 100))
(tweakable '*max-class* 20 ;5
  "Maximum number of classifications to retrieve for an access action"
  '(integer 0 100))

(defun add-access-actions (s-item &aux
				  (wff-name (subgoal-subgoal s-item))
				  int lc lit con cl access-item)
  (setq lc (getp wff-name 'LIT-CLASSES))
  (unless lc
	  (classify wff-name)
	  (setq lc (getp wff-name 'LIT-CLASSES))
	  )
  (remprop wff-name 'LIT-CLASSES)
  (dolist (l lc)
	  (setq lit (car l))
	  (setq con (cadr l))
	  (setq cl (caddr l))
	  (setq int (cond ((cadddr l)) (t 0)))
	  (when cl 
		(setq cl (sort-classes cl t))
		(setq cl (add-flags cl))
		(setq access-item (make-access-info :literal lit
						    :context con
						    :interest int
						    :inherit-interest 
						    (subgoal-inherit-interest s-item)
						    :classes cl))
		(add-agenda-access s-item access-item)
	  ))
  (when (member (getp wff-name 'wff-type) '(episodic prefix))
	(let ((mclasses (mp-classify wff-name))
	      (wint (wff-interest wff-name)))
	     (setq mclasses (add-flags (sort-classes mclasses t)))
	     (setq access-item (make-access-info :literal wff-name
						 :context '+
						 :interest (- wint 10)
						 :inherit-interest 
						 (subgoal-inherit-interest s-item)
						 :mp t
						 :classes mclasses))
	     (add-agenda-access s-item access-item)
	     ))
  )


; add-agenda-access creates an agenda item to access wffs under class
; for phi, and puts the action on the agenda.

(defun add-agenda-access (s-item access-item &aux aitem)
  (setq aitem (make-agenda-item :action-type 'access 
				:access-action access-item
				:complexity (subgoal-complexity s-item)
				:interest (access-info-interest access-item)
				:inherit-interest (access-info-inherit-interest access-item)
				:prob (subgoal-prob s-item)
				:c-prob (subgoal-c-prob s-item)
				:subgoal-parent s-item ))
  (setf (subgoal-access-items s-item) (append (subgoal-access-items s-item) 
					      (list aitem)))
  (add-agenda aitem)
  )

(defun extract-instances (wff &aux
			      (pred (getp wff 'pred))) 
  (cond ((null pred) nil)
	((numberp pred) nil)
	((and (listp pred) (numberp (car pred)) (<= (car pred) 1.0)) nil)
	((member pred '(or and implies ~> <=>)) nil)
	((member pred '(** * @ _))
	 (extract-instances (car (getp wff 'arguments))))
	(t (car (getp wff 'arguments))))
  )


; add-subgoals goes through a list of partially filled in subgoal accesses
; returned by compare-wffs, fills them out, and adds them to the agenda.

(defun add-subgoals (s-item key context kn-wff kn-wff-key kn-wff-context res 
			    int keys 
			    &aux sinfo aitem rpart)
  (setq keys (remove-if #'(lambda (a) (not (equal (cadr a) kn-wff-context)))
			keys))
  (dolist (r res)
	  (setq sinfo (make-subgoal-info :key key 
					 :context context
					 :kn-wff kn-wff
					 :kn-wff-key kn-wff-key
					 :kn-wff-context kn-wff-context
					 :comparison r))
	  (setq aitem (make-agenda-item :subgoal-parent s-item
					:action-type 'subgoal
					:subgoal-action sinfo))
	  (setf (subgoal-agenda s-item) (append (subgoal-agenda s-item)
						(list aitem)))
	  (setq rpart (cons kn-wff keys)) 
	  (setf (agenda-item-rules aitem)
		(if (or (not (getp kn-wff 'conditional)) 
			(member rpart (subgoal-rules s-item) :test 'equal))
		    (subgoal-rules s-item)
		    (cons rpart (subgoal-rules s-item) )))
	  ; disallowed forces the system to apply a non-conditional
	  ; before applying the same rule again (prevents explosions with
          ; transitive rules - if the problem is solvable, it needs to find
	  ; the instances making it true/false anyway - this just forces it
	  ; to find them before digging into a recursive hole)
	  (setf (agenda-item-disallowed-rules aitem)
		(if (getp kn-wff 'conditional)
		    (cons kn-wff (subgoal-disallowed-rules s-item))
		    nil))
	  (setf (agenda-item-prob aitem)
			       (if (and (numberp (getp kn-wff 'prob))
					(not (getp kn-wff 'assumption)))
				   (list kn-wff)
				   (getp kn-wff 'prob))
		)
	  (setf (agenda-item-c-prob aitem)
		(calc-prob-list (combine-probs (subgoal-c-prob s-item)
			       (if (and (numberp (getp kn-wff 'prob))
					(not (getp kn-wff 'assumption)))
				   (list kn-wff)
				   (getp kn-wff 'prob)))) )
	  (setf (agenda-item-complexity aitem)
		(/ (+ (- (subgoal-complexity s-item) (wff-complexity key))
		      (if (comparison-info-residue1 r) 100 0)
		      (- (getp kn-wff 'COMPLEXITY) (wff-complexity kn-wff-key))
		      (if (comparison-info-residue2 r) 100 0)) 2))
	  (setf (agenda-item-interest aitem) (cond (int) (t 0)))
	  (setf (agenda-item-inherit-interest aitem) 
		(subgoal-inherit-interest s-item))
	  (add-agenda aitem))
  )

(defun from-split (main tester &aux
			(prob (getp tester 'PROB)))
  (cond ((or (null prob) (not (listp prob))) nil)
	((> (list-length prob) 1) nil)
	((member main prob :test 'equal) t)
	(t (from-split main (car prob)))
	))

; access does an access item on the agenda. Since partial accesses are 
; allowed, this access may be new (in which case the appropriate wffs are
; retrieved from the knowledge base), or continued, in which case we have
; already retrieved the clauses and we just continue on where we left off.
; Retrieved clauses are compared against the given set-of-support clause.
; If there are any matches, subgoal actions are added to the agenda for it,
; and we place the rest of the clauses on this access item and reposition
; it on the agenda.
; When an access has finished, super or sub access actions are added as
; appropriate.

(defun access (item &aux
		    (subgoal (agenda-item-subgoal-parent item))
		    (proof-prop (subgoal-proof-prop subgoal))
		    (phi (subgoal-subgoal subgoal))
		    (access (agenda-item-access-action item))
		    (lit1 (access-info-literal access))
		    (context1 (access-info-context access))
		    (classes (access-info-classes access))
		    (current-class (access-info-current-class access))
		    (c-class (car current-class))
		    (flags (cdr current-class))
		    (got (access-info-wff-list access))
		    (already (access-info-wffs-tried access))
		    (rules (subgoal-rules subgoal)) 
		    (disallowed-rules (subgoal-disallowed-rules subgoal)) 
		    (acc-already (access-info-classes-tried access))
		    (int (access-info-interest access))
		    (num-wffs 0)
		    (num-classes 0)
		    (rest got)
		    m already-keys match wff keys)

  (when (and (traced-p 'qa-access) got)
	(print-line "Continuing access for " 
		       (if (equal proof-prop 'PROOF) 
			   "PROOF " "DISPROOF ")
		       "clause " (getp phi  'print)
		       *nl* *spaces* "     for literal " (getp lit1 'print)
		       ;"    Depth: " (subgoal-depth subgoal)
		       *nl* *spaces* "        using classification " c-class)
	)

  ; continue comparing against wffs retrieved until match found,
  ; max num of wffs reached, or max num of classes reached

  (do ()
      ((or match		            ; match found with a wff
	   (>= num-wffs *max-wffs*)         ;> max number of wffs tested
	   (>= num-classes *max-class*)     ;> max number of classes retrieved
	   (and (null got) (null classes))))  ; nothing left to look for

      ; if nothing to compare against, retrieve some using next class
    (unless got 
      (when (equal (getp phi 'wff-type) 'quantified)
	;; try getting subclasses first
	(setq classes (early-sub classes))
	(setf (access-info-classes access) classes)
	)
      
      (setq num-classes (+ num-classes 1))
      (setq current-class (car classes))
      (setq classes (cdr classes))
      (setq c-class (car current-class))
      (setq flags (cdr current-class))
      (do ()
	  ((or (null c-class) (not (member c-class acc-already :test 'equal))))
	(setq current-class (car classes))
	(setq classes (cdr classes))
	(setq c-class (car current-class))
	(setq flags (cdr current-class))
	)
	  (when c-class
	    (pushnew c-class acc-already)
	    (when (traced-p 'qa-access)
		(print-line "Access for "
			    (if (equal proof-prop 'PROOF) "PROOF " "DISPROOF ")
			    "clause " (getp phi 'print)
			    *nl* *spaces* "     for literal " (getp lit1 'print)
			    ;;"    Depth: " (subgoal-rank subgoal)
			    *nl* *spaces* "        using classification " c-class)
		)
	    (setf (access-info-classes-tried access) acc-already)
	    (setf (access-info-current-class access) current-class)
	    (setf (access-info-classes access) classes)
	    (setq got (get-class c-class))
	    (setq rest got)) )
    

  ; go through list of retrieved wffs, looking for one that will match
  ; phi. When found, add appropriate subgoal actions, and put the rest of
  ; them back on the agenda.

   (when got
         (setq wff (car got))
         (setq got (cdr got))
	 (setq rest (cdr rest))
	  ; if this is a proof access, ignore all disproof clauses retrieved
	  ; and vice versaa
	  (unless (or ;;(equal wff phi)
		   ;; I have no idea why the above was here.  If the
		   ;; retrieved wff is identical to the subgoal, then
		   ;; we should be done!  -ANK 9/2/99
		      (member wff already)
		      (if (getp wff 'assumption)
			  (not (member wff (subgoal-assumptions subgoal))))
		      (and (< *question-effort* 2) (getp wff 'conditional))
		      (member wff disallowed-rules)
		      )

		  (when (traced-p 'qa-access)
			(print-line "Retrieved " (getp wff 'print)
			    *nl* *spaces* "           to compare against " 
			    (getp lit1  'print) *nl* *spaces* 
			    "           in " (getp phi 'print)))
   
		  ; only include possibly usble clauses in count
		  (setq num-wffs (+ num-wffs 1))
		  ; compare retrieved wff with the set-of-support one
		  (push wff already)
		  (setq already-keys nil)
		  (setq keys (cdr (assoc wff rules)))
		  (when keys (setq already-keys t))
		  (if (null keys) 
		      (setq keys (make-keylist (unless (getp wff 'meaning-postulate)
						       (getp wff '-trigger))
					       (getp wff '+trigger)))
		      )
		  (setq match (try-wff subgoal phi lit1 context1 wff keys int))
		  (when (and (not already-keys)
			     (getp wff 'equivalence))
			(if (getp wff 'meaning-postulate)
			    (setq keys (make-keylist nil (getp wff '-trigger)))
			    (setq keys (reverse-keylist keys)))
			(setq m (try-wff subgoal phi lit1 context1 wff keys int))
			(setq match (or m match)) )

		  )
	 )
  
      ; add sub and super if nec, put item back on agenda
      (unless (or rest (null c-class))
	  (setq classes 
		(add-sub-super-accesses access c-class flags)))

      )  ;end of do

  (when (or classes rest)
  (setf (access-info-classes access) classes)
  (setf (access-info-classes-tried access) acc-already)
  (setf (access-info-wff-list access) rest)
  (setf (access-info-wffs-tried access) already)
  (if rest
      (setf (access-info-current-class access) current-class)
      (setf (access-info-current-class access) nil))
  (reposition-access item))
  (unless (or classes rest)
	  ; remove agenda item from subgoal list
	  (setf (subgoal-access-items subgoal)
		(remove item (subgoal-access-items subgoal)))
	  )

  (when (and (null classes) (null rest)
	     (member (getp phi 'quantifier) '(E WH))
	     (getp phi 'restrict))
	; effort level here? any other way to tell when useful?
	; add access item for the non-key type literals in the formula
	)
  nil)

(defun try-wff (subgoal phi lit1 context1 wff keys int 
			&aux res newlit context match
			(mp (getp wff 'meaning-postulate)))
  (dolist (l keys)
    (unless nil 
      ;; use context info
      (setq newlit (car l))
      (setq context (cadr l))
      (unless (and (type-lit-p lit1) 
		   (if (equal context '+) 
		       (not (equal (neg-p lit1) (neg-p newlit)))
		     (equal (neg-p lit1) (neg-p newlit))))
	(when (traced-p 'qa-test)
	  (print-line "Comparing " 
		      (getp lit1 'print) " " context1 " in " 
		      (subgoal-proof-prop subgoal) " subgoal "
		      (getp phi 'print) *nl* *spaces*
		      "           against " 
		      (getp newlit 'print) " " context " in " 
		      (getp wff 'print)))
	;; if the key in ther retrieved formula is -,
	;; comapre for incompatibility, else for
	;; comaptibility
	(if (equal context '-) 
	    (setq res (compare-wffs newlit lit1 nil
				    :prefer 2
				    :types1 (getp wff 'epilib::types)
				    :types2 (getp phi 'epilib::types)
				    :using-mp mp
				    :traced (traced-p 'qa-test) 
				    :assumption (getp wff 'assumption)
				    :header "      ")) 
	  (setq res (compare-wffs lit1 newlit t
				  :prefer 1
				  :types1 (getp phi 'epilib::types)
				  :types2 (getp wff 'epilib::types)
				  :using-mp mp
				  :assumption (getp wff 'assumption)
				  :traced (traced-p 'qa-test) 
				  :header "      ")) 
	  ) )
      (when res
	(setq match t)
	(add-subgoals subgoal lit1 context1 wff 
		      newlit context res int keys)
	)
      )
    )
  match)

(defun make-keylist (-keys +keys)
  (append (mapcar #'(lambda (a) (list a '-)) -keys)
	  (mapcar #'(lambda (b) (list b '+)) +keys))
  )

(defun reverse-keylist (keys)
  (mapcar #'(lambda (k) (list (car k) (if (equal (cadr k) '+) '- '+) (caddr k)))
	  keys))


; put the item back on the agenda, keeping track of what has already been done

(defun reposition-access (item)
  ; change position of item
  (setf (agenda-item-position item)
	(+ (agenda-item-position item) (/ *qa-access-weight* 2) ))
  (add-agenda item nil)
  )

(defun early-sub (classes &aux
			  (c1 (car classes))
			  (cl (car c1))
			  (flags (cdr c1))
			  newclasses inst-subs newc1)
  (when (member '-sub flags)
	(setq newc1 (cons cl (remove '-sub flags)))
	(setq classes (cons newc1 (cdr classes)))
	(setq inst-subs (get-instances-and-subs cl))
	(setq newclasses (append 
		 (sort (car inst-subs) #'(lambda (classa classb)
		  (let* ((tpreda (if (listp (car classa)) (caar classa) (car classa)))
			 (tpredb (if (listp (car classb)) (caar classb) (car classb)))
			 )
			(when (and (symbolp tpreda) (symbolp tpredb))
			      (order-instances tpreda tpredb))
			)))
		 (sort (cadr inst-subs) #'(lambda (classa classb)
		  (let* ((tpreda (if (listp (car classa)) (caar classa) (car classa)))
			 (tpredb (if (listp (car classb)) (caar classb) (car classb)))
			 )
			(when (and (symbolp tpreda) (symbolp tpredb))
			      (order-types tpreda tpredb))
			)))) )
	(setq classes (append (mapcar 'list newclasses) classes))
	)
  classes)

(defun get-instances-and-subs (cl &aux
				  slist ilist cllist)
  (dolist (s (remove cl (sub-class cl) :test 'equal))
	  (if (type-pred-p (if (listp (car s)) (caar s) (car s)))
	      (let ()
		   (setq slist (get-instances-and-subs s))
		   (setq ilist (append ilist (car slist)))
		   (setq cllist (append cllist (cadr slist)))
		   (setq cllist (append (list s) cllist))
		   )
	      (setq ilist (append ilist (list s)))
	      )
	  )
  (list ilist cllist))

(defun Oearly-sub (classes &aux
			  (c1 (car classes))
			  (cl (car c1))
			  (flags (cdr c1))
			  newc1)
  (when (member '-sub flags)
	(setq newc1 (cons cl (remove '-sub flags)))
	(setq classes (cons newc1 (cdr classes)))
	(dolist (c (sort-classes (sub-class cl)))
		(pushnew (cons c (list '-sub)) classes :test 'equal)))
  classes)

; check to see if adding super or sub actions is appropriate - if so,
; add them

(defun add-sub-super-accesses (item current-class flags &aux
				     (classes (access-info-classes item))
				)

  (setf (access-info-current-class item) nil)
  (setf (access-info-wff-list item) nil)

  ; add super or sub actions as necessary
	(when (member '-super flags)
	      (dolist (c (super-class current-class))
		      (unless (member (cons c (list '-super)) classes :test 'equal)
			      (setq classes (insert-in-list (cons c (list '-super)) classes))
		      )) )
	(when (member '-sub flags)
	      (dolist (c (sub-class current-class))
		      (setq classes (insert-in-list (cons c (list '-sub)) classes t))
		      ))
	(setf (access-info-classes item) classes)
  classes)

(defun insert-in-list (item list &optional fst &aux
			    (cl1 (car list))
			    (class (car cl1))
			    (first (car class))
			    (ifirst (caar item)))
  (cond ((null list) (list item))
	(t
  (if (listp first) (setq first (car first)))
  (if (listp ifirst) (setq ifirst (car ifirst)))
  (if (type-pred-p first)
      (if (type-pred-p ifirst)
	  (if (order-types ifirst first)
	      (append (list item) list)
	      (append list (list item)))
	  (append (list item) list))
      (append (list (car list)) (insert-in-list item (cdr list) fst))) 
	 ))
  )


