;       Last change    August 1993
(in-package string)

; Evaluation routines

; evaluate determines if the given literal is 	true, false or unknown
; it first ensures that the literal is in its domain

(defun evaluate (netname lit negp pred arglist effort &aux
			 (res 'unknown))
  (declare (ignore effort lit netname))
  (if (and (not (non-constant-arg arglist))
	   (symbolp pred)
	   (check-string-args pred arglist))
      (if (equal (symbol-package pred) (find-package 'lisp))
	  (setq res (call-lisp-string negp pred arglist))
	  (setq res (call-user-string negp pred arglist))
	  )
      )
  res)

(defun check-string-args (pred arglist)
  (case pred
	((string= string-equal string< string> string<= string>= string/=
          string-lessp string-greaterp string-not-lessp string-not-greaterp
	  string-not-equal)
	 (and (stringp (car arglist)) (stringp (cadr arglist))
	      ; and the rest may be keys
	      ))
	(otherwise (all-string-args arglist))
	))

(defun all-string-args (arglist)
  (eval (append (list 'and) (mapcar #'(lambda (a) (stringp a)) arglist)))
  )

(defun call-lisp-string (negp pred arglist &aux res)
  (setq res (apply pred arglist))
  (if negp (setq res (not res)))
  (if res (setq res 'YES) (setq res 'NO))
  (if (traced-p 'epilog::string-eval)
      (print-line "    String Specialist: Evaluated " (if negp "not" "")
		  " " (car arglist) " " pred " " (car arglist) " to " res))
  res)

(defun string-starts (stra start strb l2 &aux (res t))
  (do ((i 0 (+ i 1)))
      ((>= i l2))
      (unless (equal (aref stra (+ start i)) (aref strb i))
	      (return (setq res nil)))
      )
  res
  )

(defun substring-p (stra strb &aux
			     (l1 (length stra))
			     (l2 (length strb))
			     res)
  (do ((i 0 (+ i 1)))
      ((> i (- l1 l2)))
      (if (string-starts stra i strb l2)
	  (return (setq res t))
	  ))
  res)


(defun call-user-string (negp pred arglist &aux 
			      (res 'unknown)
			      (reln (find-symbol (symbol-name pred) 'string))
			      )
  (case reln
	(string-contain 
	 (setq res (substring-p (car arglist) (cadr arglist)))
	 (if negp (setq res (not res)))
	 (if res (setq res 'YES) (setq res 'NO))
	 res)
	)

  (if (and (traced-p 'epilog::string-eval) (not (equal res 'unknown)))
      (print-line "    String Specialist: Evaluated " (if negp "not" "") " "
		  (car arglist) " " pred " " (cadr arglist) " to " res))
  res
  )
