;       Last change    August 1993
(in-package string)

(tweakable 'epilog::*string-separator* nil 
  "Single character to automatically insert between every pair of strings when using string-concat"
  '(or character (member nil)))
(tweakable 'epilog::*string-field-divider* '(#\-)
  "A list of characters which can be used to determine field separation when using string-field"
  'character-list)

; determines the substring portion of s desired - s is assumed to be a string
; separated by -

(defun string-part (s num  &aux
		      (index 0)
		      (start 0)
		      part stop (num-div 0)
		      (num-chars (length s)))
  (do ()
      ((or part stop))
      (when (or (>= index num-chars) 
		(member (geta s index) epilog::*string-field-divider*))
	    (setq num-div (+ num-div 1))
	    (if (equal num-div num)
		(setq part (substring s start (- index 1)))
		(if (>= index num-chars)
		    (setq stop t)
		    (setq start (+ index 1))
		    )
		)
	    )
      (setq index (+ index 1))
      )
  part
  )

;substring returns a new string consisting of the elements between start
; and end in the old string, endpoints included

(defun substring (s start end &aux news)
  (setq news (make-string (+ (- end start) 1)))
  (do ((i start (+ i 1))
       (i2 0 (+ i2 1))) 
      ((> i end)) 
      (setf (aref news i2) (aref s i)) 
      )
  news
  )

(defun concat-strings (ss &aux
			  (res (car ss)))
  (dolist (s (cdr ss))
	  (setq res (concat-2-strings res s))
	  )
  res)

(defun concat-2-strings (s1 s2 &aux
			    res
			    (index 0)
			    (l1 (length s1))
			    (l2 (length s2))
			    )
  (setq res (make-string (+ l1 l2 (if epilog::*string-separator* 1 0))))
  (do ((i 0 (+ i 1)))
      ((>= i l1))
      (setf (aref res index) (aref s1 i))
      (setq index (+ index 1)))
  (when epilog::*string-separator*
	(setf (aref res index) epilog::*string-separator*)
	(setq index (+ index 1))
	)
  (do ((i 0 (+ i 1)))
      ((>= i l2))
      (setf (aref res index) (aref s2 i))
      (setq index (+ index 1)))
  res)


(defun string-field (&rest args &aux res
			   (str (car args))
			   (partnum (cadr args)))
  (setq res (string-part str partnum))
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::string-function-eval))
      (print-line "    String Specialist: STRING-FIELD " args
						 " evaluated to " res))
    res)

(defun sub-string (&rest args &aux res
			   (str (car args))
			   (start (cadr args))
			   (end (caddr args)))
  (if (null end) (setq end (length str)))
  (setq res (substring str start end))
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::string-function-eval))
      (print-line "    String Specialist: SUB-STRING " args
						 " evaluated to " res))
    res)

(defun string-concat (&rest args &aux res)
  (setq res (concat-strings args))
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::string-function-eval))
      (print-line "    String Specialist: STRING-CONCAT " args
						 " evaluated to " res))
    res)

(defun make-s-number (s) 
  (do ((j 0 (+ j 1))
       (n 0 (+ (* n 10)
	       (or (digit-char-p (char s j) 10)
		   (return nil)))))
      ((= j (length s)) n))
  )

(defun make-a-number (s &aux s1 s2 n1 (n2 0)
			    (save-sep epilog::*string-field-divider*))
  (setq epilog::*string-field-divider* '(#\.))
  (setq s1 (string-part s 1))
  (setq s2 (string-part s 2))
  (setq epilog::*string-field-divider* save-sep)
  (setq n1 (make-s-number s1))
  (when s2
	(setq n2 (make-s-number s2))
	(do ()
	    ((< n2 1.0))
	    (setq n2 (/ n2 10.0)))
	)
  (+ n1 n2))


(defun string-number (&rest args &aux res)
  (setq res (make-a-number (car args)))
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::string-function-eval))
      (print-line "    String Specialist: STRING-NUMBER " args
						 " evaluated to " res))
    res)


(defun number-string (&rest args &aux
			    (a1 (car args))
			    (dec-places (cadr args))
			    (total-places (caddr args))
			    anum bnum stra strb)
  (setq anum (truncate a1))
  (setq bnum (rem a1 1))

  ; convert both parts to strings
  (unless (and (equal bnum 0) (or (null dec-places) (equal dec-places 0)))
	  ; part after the decimal point
	  (do () 
	      ((equal (rem bnum 1) 0.0))
	      (setq bnum (* bnum 10)))
	  (setq bnum (truncate bnum))
	  ; pad after it with 0's to total # dec places
	  (setq strb (n-str bnum))
	  (if (and dec-places (< (length strb) dec-places))
	      (setq strb (concat-2-strings strb 
		   (make-string (- dec-places (length strb)) (digit-char 0)))))
	  (if (and dec-places (> (length strb) dec-places))
	      (setq strb (substring strb 0 (- dec-places 1))))
	  )
  ; part before
  (setq stra (n-str anum))
  (when strb
	(setq epilog::*string-separator* #\.)
	(setq stra (concat-2-strings stra strb)))
  (setq epilog::*string-separator* nil)
  (if (and total-places (< (length stra) total-places))
      ; pad with 0's in front
      (setq stra (concat-2-strings 
		   (make-string (- total-places (length stra)) 
				:initial-element (digit-char 0))
		   stra)))
  (if (and total-places (> (length stra) total-places))
      (setq stra (substring stra 0 (- total-places 1))))

  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::string-function-eval))
      (print-line "    String Specialist: NUMBER-STRING " args
						 " evaluated to " stra))
  stra)

(defun n-str (num &aux
		str index chars)
  (do* ((n num (truncate (/ n 10)))
	(m (mod n 10) (mod n 10)))
      (nil)
      (setq chars (push (digit-char m) chars))
      (if (< n 10)
	  (return))
      )
  (setq str (make-string (list-length chars)))
  (setq index 0)
  (dolist (c chars)
	  (seta str index c)
	  (setq index (+ index 1))
	  )
  str)

