;       Last change    August 1993
(in-package time)
;
(defun print-time-list (tl from)
  (dolist (item tl)
	  (print-line #\tab #\tab 
          (if (null from)
              (time-link-to-point-name item)
              (time-link-from-point-name item)
	      )) 
	  )
   )

(defun print-t-info (item) (if (null item) 'unknown item))

(defun print-time (brief key node)
  (unless (null node)
  (print-line key #\tab "Node" #\tab (time-node-name node) *nl*
	      #\tab "Chain " #\tab (meta-node-chain-number
				    (time-node-chain node)) *nl*
             #\tab "Psuedo" #\tab (time-node-psuedo node) *nl*
             #\tab "Min-psuedo" #\tab (if (null (min-psuedo node))
					  "-infinity"
					  (min-psuedo node)) *nl*
             #\tab "Max-psuedo" #\tab (if (equal (max-psuedo node) t)
					  "+infinity"
					  (max-psuedo node)) *nl*
             #\tab "Absolute-min" #\tab (print-t-info
					    (time-node-absolute-min node)) *nl*
             #\tab "Absolute-max" #\tab (print-t-info
					 (time-node-absolute-max node)) )
  (when (not brief)
        (print-line  #\tab "Ancestors")
        (print-time-list (time-node-ancestors node) t)
        (print-line  #\tab "Descendants")
        (print-time-list (time-node-descendants node) nil)
        (print-line  #\tab "XAncestors")
        (print-time-list (time-node-Xancestors node) t)
        (print-line  #\tab "XDescendants")
        (print-time-list (time-node-Xdescendants node) nil)
	)
  )
  )

(defun print-time-graph (brief tg)
  (maphash #'(lambda (key node) (print-time brief key node)) tg) )

(defun print-chain-node (brief chainnode)
  (unless (null chainnode)
	  (dolist (a (time-node-descendants chainnode))
		  (print-line #\tab "from " #\tab 
			 (time-node-name chainnode) 
			 #\tab #\tab "Psuedo"
			 #\tab (time-node-psuedo chainnode) *nl*
			 #\tab #\tab (if (time-link-strict a) "<" "<=") 
			 *nl* #\tab #\tab "Duration"
			 #\tab (print-t-info (time-link-duration-min a))
			 #\tab (print-t-info (time-link-duration-max a)) *nl*
			 #\tab "to" #\tab 
			 (time-link-to-point-name a)
			 #\tab #\tab "Psuedo" #\tab (time-node-psuedo 
						     (time-point
						      (time-link-to-point-name a))))
		  )
	  )
  (mapcar #'(lambda (d) (print-chain-node brief (time-point
						 (time-link-to-point-name d))))
	  (time-node-descendants chainnode))
  )

(defun print-meta (meta)
  (unless (null meta)
  (print-line #\tab "from" #\tab (time-link-from-point-name meta) *nl*
	     #\tab #\tab "Psuedo" #\tab (time-node-psuedo 
					 (time-point
					  (time-link-from-point-name meta)))
	     *nl*
             #\tab #\tab (if (time-link-strict meta) "<" "<=") *nl*
             #\tab #\tab "Duration"
             #\tab (print-t-info (time-link-duration-min meta))
             #\tab (print-t-info (time-link-duration-max meta)) *nl*
             #\tab "to" #\tab (time-link-to-point-name meta) *nl*
             #\tab #\tab "Chain " #\tab (meta-node-chain-number
					 (time-node-chain
					  (time-point
					   (time-link-to-point-name meta))))
	     *nl*
             #\tab #\tab "Psuedo" #\tab (time-node-psuedo
					 (time-point
					  (time-link-to-point-name meta))))
  ))

(defun print-meta-graph (brief mg &optional chain)
  (if chain
      (print-meta-node brief (geth chain mg))
    (maphash #'(lambda (key val)
		 (declare (ignore key))
		 (print-meta-node brief val))
	     mg)))

(defun print-meta-node (brief node)
  (print-line "Chain " #\tab (meta-node-chain-number node) *nl*
	      #\tab "First " #\tab (time-node-name 
				    (meta-node-first node)))
  (when (not brief)
    (mapcar 'print-meta (meta-node-connections node))
    ))

(defun print-chain (brief mg &optional chain)
  (if chain
      (progn
	(print-line "Chain " #\tab chain *nl*)
	(print-chain-node brief (meta-node-first (geth chain mg))))
    (maphash #'(lambda (key val)
		 (print-line "Chain " #\tab key *nl*)
		 (print-chain-node brief (meta-node-first val)))
	     mg)))

(defun print-net (context &aux
			  (usec context))
  (if (null context)
      (setq usec '/))
  (let ((context-array (retrieve-specialist-info *environment*
						 'time)))
    (if (null context-array)
	nil
      (geth usec context-array))))

(set-display-function 'epilog::time-info 'time-info 
  "Displays information about time points - all or one named point")
(set-display-function 'epilog::meta-info 'meta-info 
  "Displays information about links between chains - all or one named chain")
(set-display-function 'epilog::chain-info 'chain-info 
  "Displays information about links within chains - all or one named chain")
(set-display-function 'epilog::episode-info 'event-info 
  "Displays information about episodes - all or one named episode")
(set-display-function 'epilog::event-info 'event-info 
  "Same as episode-info")


(defun time-info (brief &rest args &aux 
		       (point (car args))
		       (context (cadr args))
		       (net (print-net context)))
  (when (not (null net))
	(if (or (null point)
		(equal point 'epilog::all))
	    (let ()
	         (print-line "Context" #\tab (if (null context) '/ context)
			    *nl* "Timegraph")
                 (print-time-graph brief (time-net-time-graph net))
		 )
	    (print-time brief point (time-point point))
	    )
	)
  )

(defun meta-info (brief &rest args &aux
		       (chain (car args))
		       (context (cadr args))
		       (net (print-net context)))
  (when (not (null net))
	(if (or (null chain)
		(equal chain 'epilog::all))
	    (let ()
	         (print-line "Context" #\tab (if (null context) '/ context)
			    *nl* "Metagraph")
                 (print-meta-graph brief (time-net-meta-graph net))
	         )
	    (print-meta-graph brief (time-net-meta-graph net) chain)
	    )
	)
  )

(defun chain-info (brief &rest args &aux
			 (chain (car args))
			 (context (cadr args))
			 (net (print-net context)))
  (when (not (null net))
	(if (or (null chain)
		(equal chain 'epilog::all))
	    (let ()
	         (print-line  "Context" #\tab (if (null context) '/ context)
			    *nl* "Chains")
                 (print-chain brief (time-net-meta-graph net))
	         )
	    (print-chain brief (time-net-meta-graph net) chain)
	    )
	)
  )


(defun print-event (brief key node &aux
			  (start (if node (event-node-start node)))
			  (end (if node (event-node-end node))))
  (unless (null node)
  (print-line  key *nl* " Start" #\tab start)
  (when (not brief)
	(print-time t start (time-point start))
	(terpri))
  (print-line " End" #\tab end)
  (when (not brief)
        (print-time t end (time-point end))
        (terpri))
  )
  )

(defun print-event-table (brief et)
  (maphash #'(lambda (key node) (print-event brief key node)) et)
)

(defun event-info (brief &rest args &aux
			(event (car args))
			(context (cadr args))
		        (net (print-net context)))
  (when (not (null net))
	(if (or (null event)
		(equal event 'epilog::all))
	    (let ()
	         (print-line "Context" #\tab (if (null context) '/ context)
			    *nl* "Event Table")
                 (print-event-table brief (time-net-event-table net))
		 )
	    (print-event brief event (event-point event))
	    )
	)
  )
