;       Last change    August 1993
(in-package user)
(eval-when (compile load eval)   
(provide 'time-specialist)
(require 'epilib "epilog:epilib;epilib")
(require 'spec "epilog:spec;spec")
(defpackage time-specialist (:nicknames time) (:use lisp user spec lib epilib)
  (:export
	  ;; Functions for the specialist interface

	  enter evaluate compatible-lits incompatible-lits 

	  ;; functions which may be used in propositions

	  start-of end-of date relation elapsed duration-of

	  ;; display functions

	  event-info meta-info time-info

	  ))
(unless (find-package 'time) (make-package 'time)))
(in-package time)



(traceable 'epilog::time-entry "Traces input of temporal relations")
(traceable 'epilog::time-point-entry "Traces input of point relations")
(traceable 'epilog::time-eval "Traces evaluation of temporal relations")
(traceable 'epilog::time-point-eval "Traces evaluation of point relations")
(traceable 'epilog::abs-time-entry "Traces entry of dates")
(traceable 'epilog::abs-time-eval "Traces evaluation of date relationships")
(traceable 'epilog::time-duration-entry "Traces entry of durations")
(traceable 'epilog::time-duration-eval "Traces evaluation of durations")
(traceable 'epilog::time-search "Traces metagraph searching")
(traceable 'epilog::time-test "Traces comparisons of literals")
(traceable 'epilog::time-function-eval "Traces evaluation of time functions")
(traceable 'epilog::time-all "Traces all time operations"
     '(epilog::time-entry epilog::time-point-entry epilog::time-eval 
       epilog::time-test epilog::time-point-eval epilog::abs-time-entry 
       epilog::abs-time-eval epilog::time-duration-entry 
       epilog::time-duration-eval epilog::time-search epilog::time-function-eval
       ) )
(traceable 'epilog::time "Traces interesting time stuff"
     '(epilog::time-entry epilog::time-point-entry epilog::time-eval 
       epilog::time-test epilog::time-point-eval epilog::abs-time-entry 
       epilog::abs-time-eval epilog::time-duration-entry 
       epilog::time-duration-eval epilog::time-function-eval) )
(traceable 'epilog::time-min "Traces basic time stuff"
     '(epilog::time-entry epilog::time-eval epilog::time-test 
       epilog::time-function-eval) )

; make sure these are all interned into the time package

(defconstant temporal-preds
  '(same-time equal before before-0 before-1 after after-0 after-1
    during during-0 during-1 during--0 during--1 
    during-0-0 during-0-1 during-1-0 during-1-1
    contains contains-0 contains-1 contains--0 contains--1 
    contains-0-0 contains-0-1 contains-1-0 contains-1-1
    overlaps overlaps-0 overlaps-1 overlaps--0 overlaps--1 
    overlaps-0-0 overlaps-0-1 overlaps-1-0 overlaps-1-1
    overlapped-by overlapped-by-0 overlapped-by-1 overlapped-by--0 overlapped-by--1 
    overlapped-by-0-0 overlapped-by-0-1 overlapped-by-1-0 overlapped-by-1-1
    between between-0 between-1 between--0 between--1 
    between-0-0 between-0-1 between-1-0 between-1-1
    at-most-before at-least-before exactly-before
    at-most-after at-least-after exactly-after has-duration))

(load "epilog:time;pred")
(load "epilog:time;defs")
(load "epilog:time;general")
(load "epilog:time;psuedo")
(load "epilog:time;absdur")
(load "epilog:time;links")
(load "epilog:time;find")
(load "epilog:time;eval")
(load "epilog:time;enter")
(load "epilog:time;print")
(load "epilog:time;compare")
(load "epilog:time;functions")
