;       Last change    August 1993
(in-package user)
(eval-when (compile load eval)   
(provide 'type-specialist)
(require 'lib "epilog:lib;lib")
(require 'spec "epilog:spec;spec")
(defpackage type-specialist (:nicknames type) (:use lisp user spec lib)
  (:export
	  compare-preds
	  ))
(unless (find-package 'type) (make-package 'type)))
(in-package type)

(traceable 'epilog::type-test "Shows comparisons between type predicates")

(defun compare-preds (gpred1 gpred2 &aux 
			     (pred1 (usable-pred gpred1))
			     (pred2 (usable-pred gpred2))
			    (res 'epilog::unknown))
  (when (and pred1 pred2)
        (setq res (hier-rel pred1 pred2))
	(unless (and res (not (equal res 'unknown)))
		(if (or (pred-parts pred1)
			(pred-parts pred2))
		    (setq res (hier-rel pred1 pred2))) 
		)
  (when (and (traced-p 'epilog::type-test) (not (equal res 'epilog::unknown)))
	(print-line "    Type specialist: compared " (print-info pred1) " with " 
		       (print-info pred2) " - result " res))
	)
  res)

(defun usable-pred (pred) 
  (case (pred-type pred)
	(constant pred)
	(modified
	 (if (member (getp pred 'operator)
		     '(epilog::very epilog::sort-of epilog::almost
				    epilog::extremely))
	     ; ignore operator
	     (usable-pred (car (getp pred 'arguments)))) )
	))


(defun compare-compound-preds (pred1 pred2 &aux
				     (plist1 (pred-parts pred1))
				     (plist2 (pred-parts pred2))
				     p1part1 p1part2 
				     p2part1 p2part2 rel1 rel2)
  (cond (plist1 (setq p1part1 (car plist1))
		(setq p1part2 (cadr plist1)))
	(t (setq p1part1 'entity)
	   (setq p1part2 pred1)))
  (cond (plist2 (setq p2part1 (car plist2))
		(setq p2part2 (cadr plist2)))
	(t (setq p2part1 'entity)
	   (setq p2part2 pred2)))
  ; if missing first part, always subsumed? only if known type?
  ; (e.g. is a squiggle-finger a finger?)
  (if (and p1part1 p2part1)
      (setq rel1 (hier-rel p1part1 p2part1))
      (setq rel1 (if p1part1 'subsumed 'subsumes))
      )
  (unless (or (null rel1) (equal rel1 'unknown))
	  (setq rel2 (hier-rel p1part2 p2part2)) )
  (compare-hier-res rel1 rel2))


